/*
 * lxtrap-l.a.c
 * Copyright (C) 2019, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <af.h>
#include <e4.h>
#include <e4fine.h>
#include <ln.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxlead.h>
#include <lxtrap-close.h>
#include <lxtrap-drive.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>
#include <tcvset.h>

extern const void *const x1f4_a1_walk_link;

static int line_call(void *, void *, void **);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int node_args(struct dxnews_type *, int *, unsigned,
		     struct x1f4_nodelink_type *);
static int pick_call(void *, void *, void **);
static int pick_fall(void *, void *, void **);
static int pick_roll(void *, void *, void **);
static int slip_call(void *, void *, void **);
static int slip_fall(void *, void *, void **);
static int slip_roll(void *, void *, void **);

static int
line_call(void *context, void *output, void **input)
{
    int status;
    unsigned bits;

    bits = ((struct dxnews_type *) context)->linetext_data->function.flags;
    if (bits & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	if (bits & X1f4_E4_SIDE_LIST) {
	    if (output == x1f4_a1_walk_link) {
		status = slip_roll(context, output, input);
	    } else {
		status = pick_roll(context, output, input);
	    }
	} else {
	    if (output == x1f4_a1_walk_link) {
		status = slip_fall(context, output, input);
	    } else {
		status = pick_fall(context, output, input);
	    }
	}
    } else {
	if (output == x1f4_a1_walk_link) {
	    status = slip_call(context, output, input);
	} else {
	    status = pick_call(context, output, input);
	}
    }

    return status;
}


static int
miss_call(void *text, int success, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
node_args(struct dxnews_type *line_data, int *args,
	  unsigned fold, struct x1f4_nodelink_type *nodelink_data)
{
    int status = 0;
    struct lxtrap_type *lxtrap_data;
    unsigned i;
    void **track;

    lxtrap_data = line_data->lxtrap_data;

    track = line_data->track;

    i = fold;

    do {
	int type;
	const struct lxtype_type *lxtype_data;
	unsigned miss;
	void *side;

	side = *track;

	type = *args++;

	if (type == X1f4_E4_CASE) {
	    struct x1f4_dxcast_type *dxcast_data;

	    dxcast_data = I_USER(side);
	    type = dxcast_data->type;
	    side = &dxcast_data->data;
	}

	lxtype_data = lxtrap_data->link_t.data;

	close_frame(miss, lxtype_data, lxtrap_data);

	if (miss) {
	    if (lxtype_data->bits & X1f4_LX_LINK_ACCESS) {
		status = lxtype_data->node
		    (lxtype_data->text, side, I_USER(side), nodelink_data);
		if (status) {
		    break;
		}
	    } else {
		int (*copy)(void *, void *, void **, const void *,
			    const struct x1f4_nodelink_type *);

		copy = lxtype_data->copy;
		if (copy) {
		    status = copy
			(lxtype_data->text, NULL, side, I_USER(side),
			 nodelink_data);
		    if (status) {
			break;
		    }
		}
	    }
	} else {
	}

	i--;

	track++;
    } while (i);

    line_data->fold = fold - i;

    return status;
}


static int
pick_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late;
    void **track;

    line_data = context;

    linetext_data = line_data->linetext_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    track = malloc((1 + fold + late) * sizeof(void *));
    if (track) {
	*track++ = input[-1];

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, input, late * sizeof(void *));

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);

	free(track - 1);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
pick_fall(void *context, void *output, void **input)
{
    int *args, status;
    struct dxnews_type *line_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void **used;

    used = input[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags & X1f4_E4_POST_TYPE ? 1 : 0;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    size = fold + late;

    args = malloc(size * sizeof(int) << post);
    if (args) {
	struct x1f4_function_type function;

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    memcpy(args + size, linetext_data->function.args + size,
		   size * sizeof(int));
	}

	function = linetext_data->function;

	function.args = args;
	function.count = size;
	function.flags &= ~X1f4_E4_SLIP_LIST;

	used[X1f4_E4_RENDITION] = &function;

	status = pick_call(context, output, input);

	free(args);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
pick_roll(void *context, void *output, void **input)
{
    int *args, status;
    struct dxnews_type *line_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void **used, **track;

    used = input[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    track = malloc
	((size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (track) {
	struct x1f4_function_type function;

	*track++ = input[-1];

	args = (void *) (track + size);

	function = linetext_data->function;

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    int *very;

	    very = args + size;
	    memset(very, 0, fold * sizeof(int));
	    memcpy(very + fold, function_data->args + late,
		   late * sizeof(int));

	    function.flags |= X1f4_E4_POST_TYPE;
	}

	function.args = args;
	function.count = size;
	function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	used[X1f4_E4_RENDITION] = &function;

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, input, late * sizeof(void *));

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);

	free(track - 1);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
slip_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    unsigned fold, late, size;
    void *patch;

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + (size + 1) * sizeof(void *));
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_linetext_type *linetext_data;
	void **phase, **track;

	linetext_data = line_data->linetext_data;

	dispatch_data = patch;

	track = (void *) (dispatch_data + 1);

	phase = input[1];

	*track++ = phase[-1];

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, phase, late * sizeof(void *));

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


static int
slip_fall(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void *patch, **phase, **used;

    phase = input[1];

    used = phase[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags & X1f4_E4_POST_TYPE ? 1 : 0;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + sizeof(struct x1f4_function_type)
	 + (size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (status) {
	status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
    } else {
	int *args;
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_slip;
	void **track;

	dispatch_data = patch;

	function_slip = (void *) (dispatch_data + 1);

	track = (void *) (function_slip + 1);

	args = (void *) (track + size + 1);

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    memcpy(args + size, linetext_data->function.args + size,
		   size * sizeof(int));
	}

	*function_slip = linetext_data->function;

	function_slip->args = args;
	function_slip->count = size;
	function_slip->flags &= ~X1f4_E4_SLIP_LIST;

	used[X1f4_E4_RENDITION] = function_slip;

	*track++ = phase[-1];

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, phase, late * sizeof(void *));

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


static int
slip_roll(void *context, void *output, void **input)
{
    int *args, status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void *patch, **phase, **used;

    phase = input[1];

    used = phase[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + sizeof(struct x1f4_function_type)
	 + (size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (status) {
	status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_slip;
	void **track;

	dispatch_data = patch;

	function_slip = (void *) (dispatch_data + 1);

	track = (void *) (function_slip + 1);

	*track++ = used;

	args = (void *) (track + size);

	*function_slip = linetext_data->function;

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    int *very;

	    very = args + size;
	    memset(very, 0, fold * sizeof(int));
	    memcpy(very + fold, function_data->args + late,
		   late * sizeof(int));

	    function_slip->flags |= X1f4_E4_POST_TYPE;
	}

	function_slip->args = args;
	function_slip->count = size;
	function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	used[X1f4_E4_RENDITION] = function_slip;

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, phase, late * sizeof(void *));

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


int
_libx1f4i0_lxtrap_a_logique(void *context, void *output, void **input)
{
    int linked = 0, status;
    struct dxnews_type *line_data = NULL;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_leadnode_type *leadnode_data = NULL;
    struct x1f4_linetext_type *linetext_data;

    lxtrap_data = context;

    do {
	int *args, type;
	const int *such;
	struct frame_type frame;
	struct x1f4_dxcast_type *dxcast_data;
	struct x1f4_function_type function, *function_data, *function_text;
	struct x1f4_linetext_type *linetext_over;
	struct x1f4_nodetype_type *nodetype_data;
	unsigned fold, late, mill, norm, over, post;
	void *line, **track, **used;

	dxcast_data = I_USER(input[0]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_roll(lxtrap_data, "apply", 5);
	    if (1) {
		break;
	    }
	}

	linetext_data = I_USER(&dxcast_data->data);
	if (linetext_data->function.function) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_zero(lxtrap_data, "apply", 5);
	    if (1) {
		break;
	    }
	}

	used = input[-1];

	function_data = used[X1f4_E4_RENDITION];

	fold = function_data->count - 1;

	if (fold) {
	} else {
	    l_USER(dxcast_data, output);

	    status = 0;

	    break;
	}

	over = linetext_data->function.flags;

	mill = linetext_data->function.count;

	late = mill - fold;
	if ((int) late < 0) {
	    late = 0;
	    norm = mill;
	} else {
	    norm = fold;
	}

	such = linetext_data->function.args;

	if (over & X1f4_E4_POST_TYPE) {
	    const int *pass;
	    unsigned j;

	    pass = such + mill;
	    j = norm;
	    for (; j; j--) {
		if (*pass++ & X1f4_E4_POST_XSET) {
		    status = _libx1f4i0_lxcall_stat_fink
			(&lxtrap_data->link_e, norm - j,
			 &linetext_data->function);
		    if (1) {
			break;
		    }
		}
	    }
	    if (j) {
		break;
	    }
	}

	post = over & X1f4_E4_POST_TYPE ? 1 : 0;

	status = lxtrap_data->link_m.link
	    (lxtrap_data->link_m.data, &line,
	     sizeof(struct x1f4_leadnode_type)
	     + sizeof(struct x1f4_linetext_type)
	     + sizeof(dxnews_type) + (sizeof(int) * (fold + (late << post))));
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	leadnode_data = line;

	linetext_over = (void *) (leadnode_data + 1);

	line = linetext_over + 1;

	line_data = line;

	leadnode_data->line.call = 0;

	_libx1f4i0_init_tcvset
	    (&leadnode_data->line.pset, &leadnode_data->line.text);

	line_data->lxtrap_data = lxtrap_data;

	*linetext_over = *linetext_data;

	args = (int *) (line_data + 1);

	memcpy(args, such, norm * sizeof(int));
	memcpy(args + norm, function_data->args + 1 + norm,
	       (fold - norm) * sizeof(int));

	memcpy(args + fold, such + norm, late * sizeof(int));
	if (post) {
	    memcpy(args + fold + late, such + mill + norm, late * sizeof(int));
	}

	if (over & X1f4_E4_SLIP_LIST) {
	    unsigned j;

	    j = norm;
	    for (; j; j--, args++) {
		if (*args == X1f4_E4_SLIP) {
		    *args = function_data->args[norm - j + 1];
		}
	    }

	    args -= norm;
	}

	linetext_over->context = line_data;

	linetext_over->function.args = args + fold;
	linetext_over->function.count = late;
	linetext_over->function.function = line_call;
	linetext_over->function.length = 7;
	linetext_over->function.name = "applied";
	linetext_over->function.type = linetext_data->function.type;

	linetext_over->function.flags = X1f4_E4_OVER_LINK | X1f4_E4_TEXT_LINK
	    | (over & ~(X1f4_E4_LEAD_CALL | X1f4_E4_NEWS_LINK
			| X1f4_E4_NODE_LINK));
	if (over & (X1f4_E4_LEAD_CALL | X1f4_E4_NODE_LINK)) {
	    linetext_over->function.flags |= X1f4_E4_WALK_LINK;
	}

	nodetype_data = lxtrap_data->link_x.fold[0].nodetype;

	leadnode_data->link.nodetype_data = nodetype_data;
	leadnode_data->link.slip = NULL;

	frame.data = lxtrap_data->link_m.data;
	frame.link = lxtrap_data->link_m.link;

	track = input + 1;

	function_text = (struct x1f4_function_type *) function_data;

	function.args = NULL;

	if (late) {
	    unsigned j;
	    int *half;

	    half = malloc(sizeof(int) * (1 + fold + late));
	    if (half) {
	    } else {
		status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
		if (1) {
		    break;
		}
	    }

	    function = *function_text;
	    function_text = &function;

	    function.args = half;
	    function.count = 1 + fold + late;

	    half++;
	    memcpy(half, function_data->args + 1, fold * sizeof(int));
	    half += fold;
	    j = late;
	    for (; j; j--) {
		*half++ = X1f4_E4_SLIP;
	    }
	}

	status = _libx1f4i0_lxcall_even_list
	    (&linetext_data->function, function_text, 1, &track,
	     X1f4_LXCALL_POST_COPY /* | X1f4_LXCALL_POST_WAYS */,
	     &frame, &lxtrap_data->link_p, &lxtrap_data->link_q,
	     &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data);

	if (function.args) {
	    free((void *) function.args);
	}

	if (status) {
	    line_data->track = NULL;

	    if (1) {
		break;
	    }
	}

	line_data->track = track;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_data, linetext_data,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 1;

	status = node_args(line_data, args, fold, &leadnode_data->link);
	if (status) {
	    if (1) {
		break;
	    }
	}

	status = x1f4_type_e4fine
	    (lxtrap_data->link_o.fine, &type, linetext_over);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	status = x1f4_cast_node
	    (&lxtrap_data->link_v, &lxtrap_data->link_e, nodetype_data, type,
	     &linetext_over, output);
    } while (0);

    if (status) {
	if (leadnode_data) {
	    if (line_data->track) {
		if (linked) {
		    if (line_data->fold) {
			_libx1f4i0_lxtrap_free_args(line_data, NULL);
		    }

		    _x1f4_free_effect
			(NULL, linetext_data, NULL, &leadnode_data->link);
		}

		lxtrap_data->link_m.free
		    (lxtrap_data->link_m.data, line_data->track - 1);
	    }

	    lxtrap_data->link_m.free
		(lxtrap_data->link_m.data, leadnode_data);
	}
    }

    return status;
}
